﻿using System;
using System.Collections.Generic;

using PT.Scheduler;
using PT.Scheduler.Simulation.Customizations;
using PT.SchedulerDefinitions;
using PT.Transmissions;

namespace PlanetTogether.PlanetTogether.NoAnchorDrift
{
    public class NoAnchorDrift : SchedulabilityCustomization
    {
        private bool m_optimizeIsCTP;
        private readonly ScheduledBlockManager m_blockManager = new ScheduledBlockManager();
        private readonly ATPManager m_atpManager = new ATPManager();
        readonly Dictionary<BaseId, BaseId> m_itemIdHash = new Dictionary<BaseId, BaseId>();
        readonly HashSet<BaseId> m_newOpIdsHash = new HashSet<BaseId>();
        private readonly Dictionary<long, DateTime> m_originalHoldTimes = new Dictionary<long, DateTime>(); 

        public void SimulationInitializationExternalCall(ScenarioDetail a_sd, SchedulabilitySimulationInitializationHelper a_schedulabilityHelper, ScenarioDetail.SimulationType a_simulationType, ScenarioBaseT a_transmission)
        {
            SimulationInitialization(a_sd, a_schedulabilityHelper, a_simulationType, a_transmission);
        }

        /// <summary>
        /// This function is called at the start of simulation and give you to opportunity to perform initialization.
        /// </summary>
        protected override void SimulationInitialization(ScenarioDetail a_sd, SchedulabilitySimulationInitializationHelper a_schedulabilityHelper, ScenarioDetail.SimulationType a_simulationType, ScenarioBaseT a_transmission)
        {
            m_optimizeIsCTP = false;
            if (a_simulationType != ScenarioDetail.SimulationType.Expedite && a_simulationType != ScenarioDetail.SimulationType.Optimize)
            {
                return;
            }

            List<BaseId> jobs = new List<BaseId>();
            if (a_transmission is ScenarioDetailExpediteMOsT)
            {
                ScenarioDetailExpediteMOsT expediteT = a_transmission as ScenarioDetailExpediteMOsT;
                MOKeyList.Node moNode = expediteT.MOs.First;
                while (moNode != null)
                {
                    Job job = a_sd.JobManager.GetById(moNode.Data.JobId);
                    if (job != null)
                    {
                        if (!jobs.Contains(job.Id))
                        {
                            jobs.Add(job.Id);
                        }
                    }
                    moNode = moNode.Next;
                }
            }
            else if (a_transmission is ScenarioDetailOptimizeT)
            {
                //Get all unscheduled jobs
                foreach (Job job in a_sd.JobManager)
                {
                    if (!job.Template && !job.DoNotSchedule)
                    {
                        jobs.Add(job.Id);
                    }
                }
            }

            if (jobs.Count == 0)
            {
                return;
            }

            m_optimizeIsCTP = true;

            m_blockManager.Clear();
            //m_atpManager.Clear(); //Must do in post simulation customization point because adjustments are deleted before initialization point.
            m_itemIdHash.Clear();
            m_newOpIdsHash.Clear();
            m_originalHoldTimes.Clear();

            //Store every block start and end time.
            List<Resource> resources = a_sd.PlantManager.GetResourceList();
            foreach (Resource r in resources)
            {
                if (r.Active)
                {
                    if (r.Blocks != null)
                    {
                        ResourceBlockList.Node nextBlock = r.Blocks.First;
                        while (nextBlock != null)
                        {
                            if (nextBlock.Data != null)
                            {
                                var act = nextBlock.Data.Batch.FirstActivity;
                                if (act.Anchored)
                                {
                                    ScheduledBlockInfo blockInfo = new ScheduledBlockInfo(r.Id, act);
                                    m_blockManager.AddBlock(blockInfo, r.CapacityType == InternalResourceDefs.capacityTypes.MultiTasking);
                                }
                            }
                            nextBlock = nextBlock.Next;
                        }
                    }
                }
            }

            //Store every adjustment for inventory relating to new jobs
            //Hold every operation to prevent it from moving earlier
            //foreach (BaseId newJobId in jobs)
            //{
            //    Job job = a_sd.JobManager.GetById(newJobId);
            //    List<BaseOperation> baseOperations = job.GetOperations();
            //    foreach (BaseOperation baseOperation in baseOperations)
            //    {
            //        //Keep track of each new operation to schedule
            //        m_newOpIdsHash.Add(baseOperation.Id);
            //        for (int mrI = 0; mrI < baseOperation.MaterialRequirements.Count; mrI++)
            //        {
            //            MaterialRequirement mr = baseOperation.MaterialRequirements[mrI];
            //            if (mr.ConstraintType == MaterialRequirementDefs.constraintTypes.NonConstraint) continue; // no need to track NonConstraint MRs
            //            if (mr.Warehouse == null || mr.Item == null)
            //            {
            //                throw new PT.APSCommon.PTValidationException("CTP Customization: Material Requirements must specify Item and Warehouse");
            //            }
            //            var inv = mr.Warehouse.Inventories[mr.Item.Id];
            //            if (inv == null) continue;
            //            ATPInfo atpInfo = new ATPInfo(inv);
            //            m_atpManager.AddATPInfo(atpInfo);
            //        }
            //    }
            //}

            if (a_transmission is ScenarioDetailExpediteMOsT)
            {
                foreach (Job job in a_sd.JobManager)
                {
                    List<BaseOperation> baseOperations = job.GetOperations();
                    foreach (BaseOperation op in baseOperations)
                    {
                        if (op.Scheduled)
                        {
                            BaseOperationValueChanges opChanges = new BaseOperationValueChanges(op);
                            if (!op.OnHold)
                            {
                                opChanges.OnHold = true;
                            }
                            else
                            {
                                m_originalHoldTimes.Add(op.Id.Value, op.HoldUntil);
                            }
                            opChanges.HoldUntil = op.ScheduledStartDate.Ticks;
                            a_schedulabilityHelper.SchedulabilityChanger.AddOperationChanger(opChanges);
                        }
                    }
                }
            }

            foreach (Warehouse warehouse in a_sd.WarehouseManager)
            {
                foreach (InventoryManager.ItemInventory inventory in warehouse.Inventories)
                {
                    if (m_itemIdHash.ContainsKey(inventory.Inventory.Item.Id))
                    {
                        ATPInfo atpInfo = new ATPInfo(inventory.Inventory);
                        m_atpManager.AddATPInfo(atpInfo);
                    }
                }
            }
        }

        public bool IsSchedulableExternalCall(ScenarioDetail a_sd, ScenarioDetail.SimulationType a_simType, long a_clock, long a_simulationClock, BaseResource a_res, BaseActivity a_activity, SchedulableInfo a_schedulableInfo, out long o_postEventTime)
        {
            return IsSchedulable(a_sd, a_simType, a_clock, a_simulationClock, a_res, a_activity, a_schedulableInfo, out o_postEventTime);
        }

        protected override bool IsSchedulable(ScenarioDetail a_sd, ScenarioDetail.SimulationType a_simType, long a_clock, long a_simulationClock, BaseResource a_res, BaseActivity a_activity, SchedulableInfo a_schedulableInfo, out long o_postEventTime)
        {
            o_postEventTime = -1;
            InternalActivity act = a_activity as InternalActivity;
            if (act.Anchored) return true;
            //if (act.Operation.ManufacturingOrder.Job.Name == "PT00002" || act.Operation.ManufacturingOrder.Job.Name == "PT00003")
            //{
            //    DateTime start = PT.Common.TimeZoneAdjuster.GetDisplayTimeFromUTCTicks(a_simulationClock);
            //}
            if (act == null) return true;
            if (!m_optimizeIsCTP) return true;

            BlockInfo blockToSchedule = new BlockInfo(a_res.Id, a_activity as InternalActivity);
            o_postEventTime = m_blockManager.GetScheduleDelay(blockToSchedule, a_sd, a_schedulableInfo);
            if (o_postEventTime > 0)
            {
                return false;
            }

            if (a_simulationClock >= a_sd.GetPlanningHorizonEnd().Ticks)
            {
                //Simulation is past the planning horizon, allow it to schedule without checking materials   
                return true;
            }

            for (int mrI = 0; mrI < act.Operation.MaterialRequirements.Count; mrI++)
            {
                MaterialRequirement mr = act.Operation.MaterialRequirements[mrI];
                if (mr.ConstraintType == MaterialRequirementDefs.constraintTypes.NonConstraint) continue;
                o_postEventTime = m_atpManager.GetScheduleDelayWithATP(mr.Warehouse.Id, mr.Item.Id, a_clock, a_simulationClock, mr.TotalRequiredQty, act);
                if (o_postEventTime > 0)
                {
                    return false;
                }
            }

            return true;
        }

        internal void OperationScheduled(InternalOperation a_operation, long a_simClock)
        {
            if (!m_optimizeIsCTP)
            {
                return;
            }

            if (m_newOpIdsHash.Contains(a_operation.Id))
            {
                //We need to make an adjustment to the inventory adjustments because a new operation has scheduled.

                for (int mrI = 0; mrI < a_operation.MaterialRequirements.Count; mrI++)
                {
                    MaterialRequirement mr = a_operation.MaterialRequirements[mrI];
                    if (mr.ConstraintType == MaterialRequirementDefs.constraintTypes.NonConstraint) continue;
                    if (m_atpManager.IsManaged(mr.Warehouse.Id, mr.Item.Id))
                    {
                        var inv = mr.Warehouse.Inventories[mr.Item.Id];
                        if (inv == null) continue;
                        m_atpManager.InventoryUpdate(mr.Warehouse.Id, mr.Item.Id, a_operation.ScheduledStartDate, -mr.TotalRequiredQty, a_operation);
                    }
                }

                for (int mrI = 0; mrI < a_operation.Products.Count; mrI++)
                {
                    Product product = a_operation.Products[mrI];
                    if (m_atpManager.IsManaged(product.Warehouse.Id, product.Inventory.Item.Id))
                    {
                        m_atpManager.InventoryUpdate(product.Warehouse.Id, product.Inventory.Item.Id, a_operation.ScheduledEndDate, product.TotalOutputQty, a_operation);
                    }
                }
            }
        }

        internal long GetOriginalHoldDate(BaseOperation a_op)
        {
            DateTime originalHold;
            if (m_originalHoldTimes.TryGetValue(a_op.Id.Value, out originalHold))
            {
                return originalHold.Ticks;
            }
            else
            {
                return long.MaxValue;
            }
        }

        internal void PostSimulationInventoryStorageClear()
        {
            m_atpManager.Clear();
        }

        internal void PostSimulationInventoryStorage(Inventory a_inv)
        {
            ATPInfo atpInfo = new ATPInfo(a_inv);
            m_atpManager.AddATPInfo(atpInfo);
        }

        internal bool IsCTPSimulation()
        {
            return m_optimizeIsCTP;
        }

        public override string Name => "NoAnchorDrift";

        public override string Description => "Prevents anchored activities from drifting during optimize";
    }
}
