﻿using System;
using System.Collections.Generic;

using PT.APSCommon.Range;
using PT.Common.Range;
using PT.Scheduler;

namespace PlanetTogether.PlanetTogether.NoAnchorDrift
{
    class ScheduledBlockManager
    {
        readonly Dictionary<BaseId, List<ScheduledBlockInfo>> m_resourceBlocks;
        readonly Dictionary<BaseId, List<ScheduledBlockInfo>> m_resourceBlocksMulti;

        public ScheduledBlockManager ()
        {
            m_resourceBlocks = new Dictionary<BaseId, List<ScheduledBlockInfo>>();
            m_resourceBlocksMulti = new Dictionary<BaseId, List<ScheduledBlockInfo>>();
        }

        public void AddBlock(ScheduledBlockInfo a_blockInfo, bool a_multiTasking)
        {
            List<ScheduledBlockInfo> blockList;
            if (a_multiTasking)
            {
                if (m_resourceBlocksMulti.TryGetValue(a_blockInfo.ResourceId, out blockList))
                {
                    blockList.Add(a_blockInfo);
                }
                else
                {
                    blockList = new List<ScheduledBlockInfo>();
                    blockList.Add(a_blockInfo);
                    m_resourceBlocksMulti.Add(a_blockInfo.ResourceId, blockList);
                }
            }
            else
            {
                if (m_resourceBlocks.TryGetValue(a_blockInfo.ResourceId, out blockList))
                {
                    blockList.Add(a_blockInfo);
                }
                else
                {
                    blockList = new List<ScheduledBlockInfo>();
                    blockList.Add(a_blockInfo);
                    m_resourceBlocks.Add(a_blockInfo.ResourceId, blockList);
                }
            }
        }

        public void Clear()
        {
            m_resourceBlocks.Clear();
            m_resourceBlocksMulti.Clear();
        }

        public long GetScheduleDelay(BlockInfo a_blockToSchedule, ScenarioDetail a_sd, SchedulableInfo a_schedulableInfo)
        {
            List<ScheduledBlockInfo> blockList;
            if (m_resourceBlocks.TryGetValue(a_blockToSchedule.ResourceId, out blockList))
            {
                return GetScheduleDelaySingleTasking(a_blockToSchedule, blockList, a_schedulableInfo.m_scheduledStartDate, a_schedulableInfo.m_finishDate);
            }
            else if (m_resourceBlocksMulti.TryGetValue(a_blockToSchedule.ResourceId, out blockList))
            {
                return GetScheduleDelayMultiTasking(a_blockToSchedule, blockList, a_schedulableInfo.m_scheduledStartDate, a_schedulableInfo.m_finishDate, a_sd);
            }
            else
            {
                //Attempting to schedule on a resource with no block data. This could be an error or the first block on the resource
            }

            return 0;
        }

        private static long GetScheduleDelaySingleTasking(BlockInfo a_blockToSchedule, List<ScheduledBlockInfo> a_blockList, long a_scheduledStart, long a_scheduledEnd)
        {
            List<int> removalList = new List<int>();
            long nextAttemptTime = 0;
            for (int blockI = 0; blockI < a_blockList.Count; blockI++)
            {
                ScheduledBlockInfo blockInfo = a_blockList[blockI];
                if (blockInfo.EndTime.Ticks < a_scheduledStart)
                {
                    //This value is in the past, it can be removed from the list for faster searches in the future.
                    removalList.Add(blockI);
                    continue;
                }
                if (a_scheduledEnd > blockInfo.StartTime.Ticks && a_scheduledEnd <= blockInfo.EndTime.Ticks
                    || a_scheduledStart >= blockInfo.StartTime.Ticks && a_scheduledStart < blockInfo.EndTime.Ticks
                    || a_scheduledStart <= blockInfo.StartTime.Ticks && a_scheduledEnd >= blockInfo.EndTime.Ticks)
                {
                    if (blockInfo.ActId == a_blockToSchedule.ActId)
                    {
                        //This block was scheduled here before. Allow it to schedule again.
                        RemovePreviousBlocks(a_blockList, removalList);
                        return 0;
                    }
                    //The block being scheduled runs during another blocks run.
                    nextAttemptTime = blockInfo.EndTime.Ticks + 1;
                    break;
                }
                else
                {
                    //Testing
                }
            }

            RemovePreviousBlocks(a_blockList, removalList);
            return nextAttemptTime;
        }

        private static long GetScheduleDelayMultiTasking(BlockInfo a_blockToSchedule, List<ScheduledBlockInfo> a_blockList, long a_scheduledStart, long a_scheduledEnd, ScenarioDetail a_sd)
        {
            //NOTE: This assumes that there are no overlapping online intervals
            List<int> removalList = new List<int>();
            DateTime scheduledStart = new DateTime(a_scheduledStart);
            DateTime scheduledEnd = new DateTime(a_scheduledEnd);

            List<ScheduledBlockInfo> narrowedBlockList = new List<ScheduledBlockInfo>();
            for (int blockI = 0; blockI < a_blockList.Count; blockI++)
            {
                ScheduledBlockInfo blockInfo = a_blockList[blockI];
                if (blockInfo.EndTime.Ticks < a_scheduledStart)
                {
                    //This value is in the past, it can be removed from the list for faster searches in the future.
                    removalList.Add(blockI);
                    continue;
                }

                if (a_scheduledEnd > blockInfo.StartTime.Ticks && a_scheduledEnd <= blockInfo.EndTime.Ticks
                    || a_scheduledStart >= blockInfo.StartTime.Ticks && a_scheduledStart < blockInfo.EndTime.Ticks
                    || a_scheduledStart <= blockInfo.StartTime.Ticks && a_scheduledEnd >= blockInfo.EndTime.Ticks)
                {
                    if (blockInfo.ActId == a_blockToSchedule.ActId)
                    {
                        //This block was scheduled here before. Allow it to schedule again.
                        RemovePreviousBlocks(a_blockList, removalList);
                        return 0;
                    }
                    else
                    {
                        narrowedBlockList.Add(blockInfo);
                    }
                }
            }


            //Find all intervals within the required scheduling duration
            List<ResourceCapacityInterval> capacityList = new List<ResourceCapacityInterval>();
            Resource resource = a_sd.PlantManager.GetResource(a_blockToSchedule.ResourceId) as Resource;

            DateTime searchDate = scheduledStart;
            //Find first online interval where the scheduling block would like to start
            ResourceCapacityInterval rci = resource.ResourceCapacityIntervals.FindFirstOnlineOrOvertimeIntervalContainingDate(searchDate);
            while (rci != null)
            {
                capacityList.Add(rci);

                //Search for the next online interval
                searchDate = rci.EndDateTime;

                if (searchDate > scheduledEnd)
                {
                    //The remaining intervals are too far in the future
                    break;
                }

                rci = resource.ResourceCapacityIntervals.FindFirstOnlineAfter(searchDate);

                if (rci == null || rci.StartDateTime > scheduledEnd)
                {
                    break;
                }
            }

            //For each interval, check if there is any point in which there is not enough capacity available.
            foreach (ResourceCapacityInterval interval in capacityList)
            {
                decimal nbrOfPeople = interval.NbrOfPeople;
                //Find the range at which the scheduling block and interval overlap
                DateTime start = new DateTime(Math.Max(a_scheduledStart, interval.StartDateTime.Ticks));
                DateTime end = new DateTime(Math.Min(a_scheduledEnd, interval.EndDateTime.Ticks));
                RangeValidator<long> rangeTester = new RangeValidator<long>();                

                foreach (ScheduledBlockInfo scheduledBlock in narrowedBlockList)
                {
                    if (end < scheduledBlock.StartTime || start > scheduledBlock.EndTime)
                    {
                        continue;
                    }
                    //Find the range at which the interval and the scheduled block overlap
                    DateTime blockStart = new DateTime(Math.Max(start.Ticks, scheduledBlock.StartTime.Ticks));
                    DateTime blockEnd = new DateTime(Math.Min(end.Ticks, scheduledBlock.EndTime.Ticks));
                    rangeTester.Add(blockStart.Ticks, blockEnd.Ticks);
                }

                //Find the number of blocks that overlap. 
                //If more than the number of people allowed on the inter, there isn't enough capacity
                long firstIntersectionPoint;
                long lastIntersectionPoint;
                int blocksOverlapped = rangeTester.FindIntersection(out firstIntersectionPoint, out lastIntersectionPoint);
                if (blocksOverlapped >= nbrOfPeople)
                {
                    //Try again after the last overlap
                    RemovePreviousBlocks(a_blockList, removalList);
                    return lastIntersectionPoint + 1;
                }
            }
        

            //There is enough capacity
            RemovePreviousBlocks(a_blockList, removalList);
            return 0;
        }

        private static void RemovePreviousBlocks(List<ScheduledBlockInfo> a_scheduledBlockList, List<int> a_removalIndexList)
        {
            a_removalIndexList.Sort();
            for (int i = a_removalIndexList.Count - 1; i >= 0; i--)
            {
                a_scheduledBlockList.RemoveAt(i);
            }
        }
    }
}

////For each block find the overlapping duration and subtract from available capacity
//foreach (ScheduledBlockInfo overlappingBlock in narrowedBlockList)
//{
//    //Calculate usage of this block
//    long start = Math.Max(overlappingBlock.ScheduledStartTime.Ticks, a_scheduledStart);
//    long end = Math.Min(overlappingBlock.EndTime.Ticks, a_scheduledEnd);

//    foreach (ResourceCapacityInterval interval in capacityList)
//    {
//        if (end < interval.StartDateTime.Ticks || start > interval.EndDateTime.Ticks)
//        {
//            continue;
//        }

//        long capacityStart = Math.Max(start, interval.StartDateTime.Ticks);
//        long capacityEnd = Math.Min(end, interval.EndDateTime.Ticks);

//        TimeSpan capacityUsed = new TimeSpan(capacityEnd - capacityStart);
//        availableCapacity = availableCapacity.Subtract(capacityUsed);
//    }

//    if (availableCapacity < a_requiredCapacity)
//    {
//        //Not enough capacity
//        capacityAvailable = false;
//        break;
//    }
//}

//if (!capacityAvailable)
//{
//    return a_scheduledStart + 1;
//}   
